@@ -16,22 +16,40 @@ package org.apache.lucene.benchmark.byTask.tasks;
  * limitations under the License.
  */
 
+import java.util.HashMap;
+import java.util.Map;
+
 import org.apache.lucene.benchmark.byTask.PerfRunData;
 import org.apache.lucene.index.IndexWriter;
 
 /**
- * Commits the IndexReader.
+ * Commits the IndexWriter.
  *
  */
 public class CommitIndexTask extends PerfTask {
+  String commitUserData = null;
+  
   public CommitIndexTask(PerfRunData runData) {
     super(runData);
   }
-
+  
+  public boolean supportsParams() {
+    return true;
+  }
+  
+  public void setParams(String params) {
+    commitUserData = params;
+  }
+  
   public int doLogic() throws Exception {
     IndexWriter iw = getRunData().getIndexWriter();
-    if (iw!=null) {
-      iw.commit();
+    if (iw != null) {
+      if (commitUserData == null) iw.commit();
+      else {
+        Map map = new HashMap();
+        map.put(OpenReaderTask.USER_DATA, commitUserData);
+        iw.commit(map);
+      }
     }
     
     return 1;
